//----------------------------------------------------------------------------
//
// File:        ti-disk.cpp
// Date:        27-Mar-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1998-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "cartridge.hpp"
#include "tms9900.hpp"
#include "device.hpp"
#include "ti-disk.hpp"

DBG_REGISTER ( __FILE__ );

extern USHORT ProgramCounter;

cDisk::cDisk ()
{
    FUNCTION_ENTRY ( this, "cDisk::cDisk", true );
    
    memset ( m_Dirty, 0, sizeof ( m_Dirty ));
    memset ( m_DiskFile, 0, sizeof ( m_DiskFile ));
    memset ( m_DiskImage, 0, sizeof ( m_DiskImage ));
    m_CurDisk = NULL;

    m_Status = 0x7E;
    m_LastData = 0x00;
    
    m_TrapIndex = ( UCHAR ) -1;
    
    m_DriveSelect = 0;
    m_HeadSelect = 0;
    m_TrackSelect = 0;
    m_TrackIndex = 0;
    m_SectorIndex = 0;
    m_TransferEnabled = false;
    m_WriteTrack = false;

    m_DataPtr = NULL;
    m_DataBytes = 0;

    m_CRU = 0x1100;
    m_pROM = new cCartridge ( "TIDISK.rom" );
}

cDisk::~cDisk ()
{
    FUNCTION_ENTRY ( this, "cDisk::~cDisk", true );
    
    UnLoadDisk ( 0 );
    UnLoadDisk ( 1 );
    UnLoadDisk ( 2 );
}

void cDisk::SaveImage ( FILE *file )
{
    FUNCTION_ENTRY ( this, "cDisk::SaveImage", true );
    
    fwrite ( &m_Status, sizeof ( m_Status ), 1, file );
    fwrite ( &m_LastData, sizeof ( m_LastData ), 1, file );
    fwrite ( &m_DriveSelect, sizeof ( m_DriveSelect ), 1, file );
    fwrite ( &m_HeadSelect, sizeof ( m_HeadSelect ), 1, file );
    fwrite ( &m_TrackSelect, sizeof ( m_TrackSelect ), 1, file );
    fwrite ( &m_TrackIndex, sizeof ( m_TrackIndex ), 1, file );
    fwrite ( &m_SectorIndex, sizeof ( m_SectorIndex ), 1, file );
    fwrite ( &m_TransferEnabled, sizeof ( m_TransferEnabled ), 1, file );
    fwrite ( &m_WriteTrack, sizeof ( m_WriteTrack ), 1, file );

    for ( unsigned i = 0; i < SIZE ( m_DiskFile ); i++ ) {
        const char *name = m_DiskFile [i];
        fputc ( name ? strlen ( name ) : 0, file );
        if ( name ) fwrite ( name, strlen ( name ), 1, file );
    }

//    sDiskInfo    *m_CurDisk;
//    sSector      *m_CurSector;

    fwrite ( &m_DataBytes, sizeof ( m_DataBytes ), 1, file );
    fwrite ( m_AddressBuffer, sizeof ( m_AddressBuffer ), 1, file );
    fwrite ( m_TrackBuffer, sizeof ( m_TrackBuffer ), 1, file );

//    BYTE         *m_DataPtr;
}

void cDisk::LoadImage ( FILE *file )
{
    FUNCTION_ENTRY ( this, "cDisk::LoadImage", true );
    
    fread ( &m_Status, sizeof ( m_Status ), 1, file );
    fread ( &m_LastData, sizeof ( m_LastData ), 1, file );
    fread ( &m_DriveSelect, sizeof ( m_DriveSelect ), 1, file );
    fread ( &m_HeadSelect, sizeof ( m_HeadSelect ), 1, file );
    fread ( &m_TrackSelect, sizeof ( m_TrackSelect ), 1, file );
    fread ( &m_TrackIndex, sizeof ( m_TrackIndex ), 1, file );
    fread ( &m_SectorIndex, sizeof ( m_SectorIndex ), 1, file );
    fread ( &m_TransferEnabled, sizeof ( m_TransferEnabled ), 1, file );
    fread ( &m_WriteTrack, sizeof ( m_WriteTrack ), 1, file );

    for ( unsigned i = 0; i < SIZE ( m_DiskFile ); i++ ) {
        if ( m_DiskFile [i] ) free (( void * ) m_DiskFile [i] );
        int length = fgetc ( file );
        if ( length ) {
            char * name = ( char * ) malloc ( length + 1 );
            fread ( name, length, 1, file );
            name [length] = '\0';
            m_DiskFile [i] = name;
        } else {
            m_DiskFile [i] = 0;
        }
    }

//    sDiskInfo    *m_CurDisk;
//    sSector      *m_CurSector;

    fread ( &m_DataBytes, sizeof ( m_DataBytes ), 1, file );
    fread ( m_AddressBuffer, sizeof ( m_AddressBuffer ), 1, file );
    fread ( m_TrackBuffer, sizeof ( m_TrackBuffer ), 1, file );

//    BYTE         *m_DataPtr;
}

void cDisk::LoadDisk ( int index, const char *filename )
{
    FUNCTION_ENTRY ( this, "cDisk::LoadDisk", true );
    
    sDiskInfo *disk = &m_DiskImage [index];
    sSectorInfo info;

    m_DiskFile [index] = strdup ( filename );
    FILE *diskFile = fopen ( filename, "rb" );
    if ( diskFile == NULL ) {
        ERROR ( "Unable to open file " << filename );
        return;
    }

    EVENT ( "Loading disk: " << m_DiskFile [index] );

    memset ( disk, 0, sizeof ( *disk ));
    int track = 0, sector = 0;
    int ch = fgetc ( diskFile );
    while ( ! feof ( diskFile )) {
        ungetc ( ch, diskFile );
        fread ( &info, sizeof ( info ), 1, diskFile );
        BYTE *ptr = info.DataCount ? new BYTE [ info.DataCount ] : NULL;
        if ( ptr ) fread ( ptr, info.DataCount, 1, diskFile );
        if ( info.ActualCylinder != track ) {
            disk->Track[track].noSectors = sector;
            track += 1;
            sector = 0;
        }
        if (( sector > 9 ) || ( track > 39 )) {
            ERROR ( "Error reading disk file" );
            return;
        }

        char buffer [256];
        sprintf ( buffer, "  T:%02X S:%02X - AC:%02X AH:%02X - C:%02X H:%02X S:%02X L:%02X - %d",
            track, sector, info.ActualCylinder, info.ActualSide, 
            info.LogicalCylinder, info.LogicalSide, info.LogicalSector,
            info.Length, info.DataCount );
        EVENT ( buffer );

        disk->Track[track].Sector[sector].Info = info;
        disk->Track[track].Sector[sector++].Data = ptr;

        ch = fgetc ( diskFile );
    }
    disk->Track[track].noSectors = sector;
    disk->noTracks = track + 1;
    fclose ( diskFile );
}
    
void cDisk::UnLoadDisk ( int index )
{
    FUNCTION_ENTRY ( this, "cDisk::UnLoadDisk", true );
    
    EVENT ( "Removing disk: " << m_DiskFile [index] );

    free (( void * ) m_DiskFile [index] );
    m_DiskFile [index] = NULL;

    sDiskInfo *disk = &m_DiskImage [index];
    for ( int t = 0; t < disk->noTracks; t++ ) {
        sTrack *track = &disk->Track [t];
        for ( int s = 0; s < track->noSectors; s++ ) {
            if ( track->Sector [s].Data ) delete track->Sector [s].Data;
        }
    }

    memset ( disk, 0, sizeof ( *disk ));
}

USHORT cDisk::TrapFunction ( void *ptr, int, bool read, const ADDRESS address, USHORT value )
{
    FUNCTION_ENTRY ( ptr, "cDisk::Trap", false );

    cDisk *pThis = ( cDisk * ) ptr;

    if ( read == true ) {
        value = pThis->ReadMemory ( address, value );
    } else {
        value = pThis->WriteMemory ( address, value );
    }

    return value;
}

void cDisk::Activate ()
{
    FUNCTION_ENTRY ( this, "cDisk::Activate", true );
    
    if ( m_TrapIndex != ( UCHAR ) -1 ) return;
    
    m_TrapIndex = m_pCPU->RegisterBreakpoint ( TrapFunction, this, TRAP_DISK );
    m_pCPU->SetBreakpoint ( 0x5FF0, ( UCHAR ) MEMFLG_READ,  true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FF2, ( UCHAR ) ( MEMFLG_READ | MEMFLG_WRITE ), true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FF4, ( UCHAR ) MEMFLG_READ,  true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FF6, ( UCHAR ) MEMFLG_READ,  true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FF8, ( UCHAR ) MEMFLG_WRITE, true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FFA, ( UCHAR ) MEMFLG_WRITE, true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FFC, ( UCHAR ) MEMFLG_WRITE, true, m_TrapIndex );
    m_pCPU->SetBreakpoint ( 0x5FFE, ( UCHAR ) MEMFLG_WRITE, true, m_TrapIndex );
}

void cDisk::DeActivate ()
{
    FUNCTION_ENTRY ( this, "cDisk::DeActivate", true );
    
    m_pCPU->DeRegisterBreakpoint ( m_TrapIndex );
    m_TrapIndex = 0;
}

void cDisk::FindSector ()
{
    FUNCTION_ENTRY ( this, "cDisk::FindSector", true );
    
    sTrack *track = &m_CurDisk->Track [ m_TrackSelect ];
    for ( int i = 0; i < track->noSectors; i++ ) {
        if (( track->Sector [i].Info.LogicalSector == m_SectorIndex ) &&
            ( track->Sector [i].Info.LogicalCylinder == m_TrackIndex )) {
            m_CurSector = &track->Sector [i];
            return;
        }
    }
    m_CurSector = NULL;
    
    ERROR (( void * ) ProgramCounter << "------ Unable to find sector " << m_SectorIndex << " on track " << m_TrackIndex << "! -------" );
}

void cDisk::Restore ()
{
    FUNCTION_ENTRY ( this, "cDisk::Restore", true );
    
    m_Status = 0x7E;
    m_TrackSelect = 0;
    m_TrackIndex = 0;
    m_SectorIndex = 0;
}

void cDisk::Seek ()
{
    FUNCTION_ENTRY ( this, "cDisk::Seek", true );
    
    m_Status = 0x00;
    m_TrackSelect = m_LastData;
    sTrack *track = &m_CurDisk->Track [ m_TrackSelect ];
    if ( m_TrackIndex != track->Sector [0].Info.LogicalCylinder ) {
#ifdef DEBUG    
        EVENT (( void * ) ProgramCounter << "Looking for Cylinder: " << hex << m_TrackIndex << " on Track " << m_TrackSelect );
        char buffer [256], *ptr = buffer;
        ptr += sprintf ( ptr, "Disk: Seek - Found Cylinders:" );
        for ( int i = 0; i < track->noSectors; i++ ) {
            if ( track->Sector [i].Info.DataCount ) {
                ptr += sprintf ( ptr, " %02X", track->Sector [i].Info.LogicalCylinder );
            }        
        }
        EVENT ( buffer );
#endif        
        m_Status |= STATUS_NOT_FOUND | 0x10;
        m_TrackIndex = track->Sector [0].Info.LogicalCylinder;
    }
}

void cDisk::StepIn ()
{
    FUNCTION_ENTRY ( this, "cDisk::StepIn", true );
    
    m_TrackSelect++;
}

UCHAR cDisk::ReadByte ()
{
    FUNCTION_ENTRY ( this, "cDisk::ReadByte", true );
    
    UCHAR retVal = 0;

    if ( m_DataBytes ) {
        m_DataBytes--;
        retVal = *m_DataPtr++;
    } else {
        ; // Set error
    }

    return retVal;
}

void cDisk::WriteByte ( UCHAR val )
{
    FUNCTION_ENTRY ( this, "cDisk::WriteByte", true );
    
    m_LastData = val;
    if ( m_DataBytes ) {
        m_DataBytes--;
        *m_DataPtr++ = val;
    } else {
        ; // Set error
        ERROR (( void * ) ProgramCounter << "Excess data: " << hex << val );
    }
}

void cDisk::ReadSector ()
{
    FUNCTION_ENTRY ( this, "cDisk::ReadSector", true );
    
    FindSector ();
    
    if ( m_CurSector != NULL ) {
        EVENT ( " T:" << hex << m_CurSector->Info.ActualCylinder	\
             << " S:" << m_CurSector->Info.ActualSide << " -"		\
             << " C:" << m_CurSector->Info.LogicalCylinder		\
             << " H:" << m_CurSector->Info.LogicalSide			\
             << " S:" << m_CurSector->Info.LogicalSector		\
             << " L:" << m_CurSector->Info.Length			\
             << " - " << dec << m_CurSector->Info.DataCount );
        m_DataBytes = m_CurSector->Info.DataCount;
        m_DataPtr = m_CurSector->Data;
        m_Status = ( UCHAR ) (( m_CurSector->Info.Length & 0x80 ) ? 0xFF : 0x00 );
    } else {
        ; // ?? set up a dummy buffer & error status ??
    }
}

void cDisk::WriteSector ()
{
    FUNCTION_ENTRY ( this, "cDisk::WriteSector", true );
    
    FindSector ();
    
    if ( m_CurSector != NULL ) {
        EVENT ( " T:" << hex << m_CurSector->Info.ActualCylinder	\
             << " S:" << m_CurSector->Info.ActualSide << " -"		\
             << " C:" << m_CurSector->Info.LogicalCylinder		\
             << " H:" << m_CurSector->Info.LogicalSide			\
             << " S:" << m_CurSector->Info.LogicalSector		\
             << " L:" << m_CurSector->Info.Length			\
             << " - " << dec << m_CurSector->Info.DataCount );

        m_Dirty [ m_CurDisk - m_DiskImage ] = true;

        m_DataBytes = m_CurSector->Info.DataCount;
        m_DataPtr = m_CurSector->Data;
        m_Status = 0x00;
    }
}

void cDisk::ReadTrack ()
{
    FUNCTION_ENTRY ( this, "cDisk::ReadTrack", true );

    memset ( m_TrackBuffer, 0xFF, sizeof ( m_TrackBuffer ));

    m_DataBytes = sizeof ( m_TrackBuffer );
    m_DataPtr = m_TrackBuffer;
    m_Status = 0x00;

    FATAL ( "Function not implemented" );
}

void cDisk::WriteTrack ()
{
    FUNCTION_ENTRY ( this, "cDisk::WriteTrack", true );
    
    m_Dirty [ m_CurDisk - m_DiskImage ] = true;

    m_WriteTrack = true;

    m_DataBytes = 0;
    m_DataPtr = NULL;
    m_Status = 0x00;

    FATAL ( "Function not implemented" );
}

void cDisk::ReadAddress ()
{
    FUNCTION_ENTRY ( this, "cDisk::ReadAddress", true );
    
    FindSector ();

    if ( m_CurSector != NULL ) {
        m_AddressBuffer [0] = m_CurSector->Info.LogicalCylinder;
        m_AddressBuffer [1] = m_CurSector->Info.LogicalSide;
        m_AddressBuffer [2] = m_CurSector->Info.LogicalSector;
        m_AddressBuffer [3] = ( UCHAR ) ( m_CurSector->Info.Length & 0x7F );
        m_AddressBuffer [4] = 0xFF;			// CRC
        m_AddressBuffer [5] = 0xFF;			// CRC

        m_DataBytes = 6;
        m_DataPtr = m_AddressBuffer;
        m_Status = 0x00;
    }
}

void cDisk::HandleCommand ( USHORT cmd )
{
    FUNCTION_ENTRY ( this, "cDisk::HandleCommand", true );

#ifdef DEBUG
    const char *txt;
    
    switch ( cmd & 0xF0 ) {
        case 0x00 :			// CMD_RESTORE
            txt = "CMD_RESTORE";	
            break;
        case 0x10 :			// CMD_SEEK
            txt = "CMD_SEEK";		
            break;
        case 0x50 :			// CMD_STEP_IN
            txt = "CMD_STEP_IN";	
            break;
        case 0x80 :			// CMD_READ
            txt = "CMD_READ";
            break;
        case 0xA0 :			// CMD_WRITE
            txt = "CMD_WRITE";		
            break;
        case 0xC0 :			// CMD_READ_ADDRESS
            txt = "CMD_READ_ADDRESS";	
            break;
        case 0xD0 :			// CMD_INTERRUPT
            txt = "CMD_INTERRUPT";	
            break;
        case 0xE0 :			// CMD_READ_TRACK
            txt = "CMD_READ_TRACK";	
            break;
        case 0xF0 :			// CMD_WRITE_TRACK
            txt = "CMD_WRITE_TRACK";	
            break;
        default :
            txt = "<Unknown>";
            return;
    }
    EVENT (( void * ) ProgramCounter << "Command = " << txt );
#endif

    switch ( cmd & 0xF0 ) {
        case 0x00 :			// CMD_RESTORE
            Restore ();
            break;
        case 0x10 :			// CMD_SEEK
            Seek ();
            break;
        case 0x50 :			// CMD_STEP_IN
            StepIn ();
            break;
        case 0x80 :			// CMD_READ
            ReadSector ();
            break;
        case 0xA0 :			// CMD_WRITE
            WriteSector ();
            break;
        case 0xC0 :			// CMD_READ_ADDRESS
            ReadAddress ();
            break;
        case 0xD0 :			// CMD_INTERRUPT
            m_Status = 0x7E;
            break;
        case 0xE0 :			// CMD_READ_TRACK
            ReadTrack ();
            break;
        case 0xF0 :			// CMD_WRITE_TRACK
            WriteTrack ();
            break;
        default :
            ERROR ( "Unknown command: " << hex << cmd );
            return;
    }
}

USHORT cDisk::WriteMemory ( const ADDRESS address, USHORT val )
{
    FUNCTION_ENTRY ( this, "cDisk::WriteMemory", true );
    
    val ^= 0xFF;
    switch ( address ) {
        case REG_COMMAND :
            HandleCommand ( val );
            break;
        case REG_WR_TRACK :
            m_TrackIndex = val;
            if ( m_TrackIndex != m_CurDisk->Track [ m_TrackSelect ].Sector [0].Info.LogicalCylinder ) {
                m_Status |= STATUS_NOT_FOUND;
            } else {
                m_Status &= ~STATUS_NOT_FOUND;
            }
            EVENT (( void * ) ProgramCounter << "Set Track = " << hex << ( UCHAR ) val );
            break;
        case REG_WR_SECTOR :
            m_SectorIndex = val;
            EVENT (( void * ) ProgramCounter << "Set Sector = " << hex << ( UCHAR ) val );
            break;
        case REG_WR_DATA :
            WriteByte (( UCHAR ) val );
            break;
        default :
            ERROR ( "*** Unexpected MEM Wr: " << hex << address << " => " << ( UCHAR ) val );
    }

    return val;
}

USHORT cDisk::ReadMemory ( const ADDRESS address, USHORT )
{
    FUNCTION_ENTRY ( this, "cDisk::ReadMemory", true );
    
    USHORT retVal = 0xFF;
    switch ( address ) {
        case REG_STATUS :
            retVal = m_Status;
            EVENT (( void * ) ProgramCounter << "Get Status = " << hex << ( UCHAR ) retVal );
            break;	 
        case REG_RD_TRACK :
            retVal = m_TrackSelect;
            EVENT (( void * ) ProgramCounter << "Get Track = " << hex << ( UCHAR ) retVal );
            break;
        case REG_RD_DATA :
            retVal = ReadByte ();
            break;
        default :
            ERROR ( "*** Unexpected MEM Rd: " << hex << address );
    }
    return ( USHORT ) ( retVal ^ 0xFF );
}

void cDisk::WriteCRU ( ADDRESS address, USHORT val )
{
    FUNCTION_ENTRY ( this, "cDisk::WriteCRU", true );
    
    switch ( address ) {
        case 1 :
            // Don't care
            break;
        case 2 :
            m_TransferEnabled = ( val != 0 ) ? true : false;
            break;
        case 3 :
            // Don't care
            break;
        case 4 :					// Bits 4,5,6
            m_DriveSelect = val;
            switch ( val ) {
                case 1 : m_CurDisk = &m_DiskImage [ 0 ];	break;
                case 2 : m_CurDisk = &m_DiskImage [ 1 ];	break;
                case 4 : m_CurDisk = &m_DiskImage [ 2 ];	break;
            }
        case 5 : 
        case 6 : 
            break;
        case 7 :
            m_HeadSelect = val;
            break;
        default:
            ERROR ( "Unexpected Address - " << address );
            break;
    }

    EVENT (( void * ) ProgramCounter << "   I/O Wr: " << hex << ( UCHAR ) address << " => " << dec << val );
}

int cDisk::ReadCRU ( ADDRESS address )
{
    FUNCTION_ENTRY ( this, "cDisk::ReadCRU", true );
    
    int retVal = 1;
    switch ( address ) {
        case 3 : 
            retVal = 0;
            break;
        case 2 : 
            retVal = 0;
            break;
        case 1 : 
            retVal = 0;
            break;
        default:
            ERROR ( "Unexpected Address - " << address );
            break;
    }

    EVENT (( void * ) ProgramCounter << "   I/O Rd: " << hex << ( UCHAR ) address << " => " << dec << retVal );

    return retVal;
}
